require "math"

local btk = require "btk"
local Blob = require "Blob"
require "Walls"
require "Bees"

-- Bees!
-- Cracks

function centerCamera(pos)
    camera = btk.Vec(pos.x - lg.getWidth() / 2, pos.y - lg.getHeight() / 2)
end

function containCamera(pos)
    local cameraCenter = camera + btk.Vec(lg.getWidth() / 2, lg.getHeight() / 2)

    if (cameraCenter - pos):mag() > 150 then
        cameraCenter = pos + (cameraCenter - pos):norm() * 150

        centerCamera(cameraCenter)
    end
end

function startGame()
    gameStartTime = lc.getTime()
    firstClickTime = nil
    worldTimeStep = 0
    beehiveHealth = 20
    worldState = "playing"
    beeQuiet = false

    resetBees()
    spawnBee(btk.Vec(44, 24))

    if blob then blob.body:destroy() end
    blob = Blob(30, 20, btk.Vec(3 * 30, 25 * 30), 1)
    blob:goCircle()
end

function printHelpMessage()
    if not firstClickTime then
        lg.setColor(255, 255, 255)
    else
        local p = (1 - ((lc.getTime() - firstClickTime) / 3)) * 255
        lg.setColor(255, 255, 255, p)
    end

    lg.setFont(font)
    lg.print([[
1. Click and drag me to fly
2. Moving very fast turns you into a triangle
3. Triangle mode does DAMAGE
4. Don't get stung by the enemy's stings

How long can you survive?]], unpack(blob.position + btk.Vec(0, 20)))

    lg.setFont(bigFont)
    lg.print("Main objective: Destroy the BEEHIVE", unpack(blob.position + btk.Vec(0, -70)))


end

function beginContact(a, b, coll)
    local aObj = fixtureMap[a]
    local bObj = fixtureMap[b]

    if aObj and bObj then
        if aObj.collidedWith then
            aObj:collidedWith(bObj, a, b)
        end
        if bObj.collidedWith then
            bObj:collidedWith(aObj, b, a)
        end
    end
end
 
function endContact(a, b, coll)
 
end
 
function preSolve(a, b, coll)
 
end
 
function postSolve(a, b, coll, normalimpulse, tangentimpulse)
 
end

function love.keypressed(key)
    if key == "1" then blob:goSquare() end
    if key == "2" then blob:goCircle() end
    if key == "3" then blob:goCompact() end
    if key == "4" then blob:goTriangle() end
    if key == "escape" then os.exit() end
    if key == "space" then worldTimeStep = 1 - worldTimeStep end
    if key == "x" then
        for _, v in ipairs(wallParticles) do
            v.body:applyLinearImpulse(love.math.random(-1000, 1000), love.math.random(-1000, 1000))
        end
    end
    if key == "b" then startGame() end
    if key == "l" then
        beeQuiet = not beeQuiet
    end

    if worldTimeStep == 0 then
        if key == "o" then
            local file = io.read()
        elseif key == "r" then
            walls:reset()
            wallParticles:reset()
        end
    end
end

local replayButtonText = "Play again?"
local replayButtonY = 300

function love.mousepressed(x, y, button)
    local left, top, right, bottom = blob.fixture:getBoundingBox()
    local aabb = btk.AABR(top, right, bottom, left)

    aabb:grow(realityPadding)

    if aabb:containsPoint(camera + btk.Vec(x, y)) and not blob.freeze then
        aiming = true
        worldTimeStep = 0.2
        if not firstClickTime then firstClickTime = lc.getTime() end
    end

    if worldState ~= "playing" then
        local width = bigFont:getWidth(replayButtonText)
        local height = bigFont:getHeight(replayButtonText)
        local aabb = btk.AABB(lg.getWidth() / 2 - width / 2,
            replayButtonY,
            width,
            height)

        if aabb:containsPoint(btk.Vec(x, y)) then
            startGame()
        end
    end
end

function love.mousereleased(x, y, button)
    if aiming and not blob.freeze then
        aiming = false

        local force = (blob.position - (camera + btk.Vec(x, y))) * 10
        blob.body:applyLinearImpulse(force.x, force.y)

        worldTimeStep = 1
    end
end

function love.load()
    if love.window then
        love.window.setMode(0, 0, {fullscreentype = "desktop"})
        love.window.setTitle("Be Gone! - Ludum Dare 35 by bobismijnnaam")
    end

    if love.system.getOS() == "Android" then realityPadding = 200
    elseif love.system.getOS() == "Linux" or love.system.getOS() == "Windows" then realityPadding = 50
    else realityPadding = 200 end

    imgs = {
        triangle = lg.newImage("triangle.png"),
        circle = lg.newImage("circle.png"),
        rectangle = lg.newImage("rectangle.png")
    }

    font = lg.newFont("Aaargh.ttf", 20)
    bigFont = lg.newFont("Aaargh.ttf", 40)
    lg.setFont(font)

    beehiveImg = lg.newImage("beehive.png")
    heartImg = lg.newImage("heart.png")
    heartImg:setFilter("nearest", "nearest")

    cloudBigImg = lg.newImage("cloudbig.png")
    cloudBigImg:setFilter("nearest", "nearest")
    cloudImg = lg.newImage("cloud.png")
    cloudImg:setFilter("nearest", "nearest")

    fixtureMap = {}

    for _, v in pairs(imgs) do
        v:setWrap("repeat", "repeat")
    end

    worldState = "playing"

    --lg.setBackgroundColor(20, 120, 210)
    lg.setBackgroundColor(120, 180, 70)

    lp.setMeter(30)
    world = lp.newWorld(0, 9.81*30, true)
    world:setCallbacks(beginContact, endContact, preSolve, postSolve)

    normalSize = 30

    -- Border walls
    walls:addWall(0, 0, 70, 1) -- top
    walls:addWall(0, 1, 1, 29) -- left
    walls:addWall(69, 1, 1, 29) -- right
    walls:addWall(0, 30, 70, 1) -- bottom

    aiming = false

    worldTimeStep = 1

    beehive = {}
    beehive.x = 40
    beehive.y = 23
    beehive.triggerType = "beehive"
    beehive.body = lp.newBody(world, beehive.x * 30 + beehiveImg:getWidth() / 2, 30 * normalSize)
    beehive.shape = lp.newPolygonShape(
        -4 * normalSize, 0,
        0, -7 * normalSize,
        4 * normalSize, 0)
    beehive.fixture = lp.newFixture(beehive.body, beehive.shape)
    fixtureMap[beehive.fixture] = beehive

    startGame()
    centerCamera(blob.position)
end

function love.update(dt)
    world:update(worldTimeStep * dt)

    if worldState == "playing" then
        if blob.health <= 0 then
            worldState = "loss"
            blob.duration = 10
            blob.goCompact()
            blob.freeze = true
            blob.diePosition = btk.clone(blob.position)
            beeQuiet = true
            stateChangeTime = lc.getTime()
        elseif beehiveHealth <= 0 then
            worldState = "win"
            woldTimeStep = 1
            blob.freeze = true
            blob.diePosition = btk.clone(blob.position)
            beeQuiet = true
            stateChangeTime = lc.getTime()
        end
    end

    blob:update(dt)

    updateBees(dt)

    if worldState == "loss" or worldState == "win" then
        containCamera(blob.diePosition)
    else
        containCamera(blob.position)
    end
end 

function love.draw()
    lg.push()

    lg.translate(-camera.x, -camera.y)

    lg.setColor(20, 120, 210)
    lg.rectangle("fill", 0, 0, 70 * 30, 30*30)

    -- Enable scissors for view
    lg.setScissor(-camera.x + 30, -camera.y, 68 * 30, 30 * 30)

    -- Draw big cloud
    lg.setColor(255, 255, 255)
    local p = ((lc.getTime() - gameStartTime) % 90) / 90
    lg.draw(cloudBigImg, (-40) * 30 + 130 * 30 * p, 6 * 30, 0, 4, 4)
    lg.draw(cloudImg, 70 * 30 - 80 * 30 * p, 10 * 30)

    -- Disable scissors
    lg.setScissor()

    walls:draw()

    if lc.getTime() - (beehiveFlash or 0) < 0.4 then
        lg.setColor(255, 0, 0)
    else
        local p = beehiveHealth / 20
        lg.setColor(255 * p, 255 * p, 255 * p)
    end
    lg.draw(beehiveImg, beehive.x * normalSize, beehive.y * normalSize)

    lg.setColor(255, 0, 0)
    lg.rectangle("fill", beehive.x * normalSize, 30 * normalSize + 10, 240, 50)
    lg.setColor(120, 180, 70)
    lg.rectangle("fill", beehive.x * normalSize + 10, 30 * normalSize + 10 + 10, 220, 30)
    lg.setColor(255, 0, 0)
    lg.rectangle("fill", beehive.x * normalSize + 10 + 10, 30 * normalSize + 10 + 10 + 10, 200 * beehiveHealth / 20, 10)

    if aiming and not blob.freeze then
        local mousePos = camera + btk.Vec(lm.getX(), lm.getY())
        local dv = blob.position - mousePos
        dv = dv:norm()

        local bottomLeft = mousePos + dv:left() * 10
        local bottomRight = mousePos + dv:right() * 10
        local topLeft = blob.position + dv:left() * 10
        local topRight = blob.position + dv:right() * 10

        local power = (blob.position - mousePos):mag() / 300
        local r = 255 * power
        local g = 255 * (1 - power)
        lg.setColor(r, g, 0)

        lg.polygon("fill",
            bottomLeft.x, bottomLeft.y,
            topLeft.x, topLeft.y,
            topRight.x, topRight.y,
            bottomRight.x, bottomRight.y)
    end

    drawBees()

    blob:draw()

    wallParticles:draw()

    printHelpMessage()

    lg.pop()

    lg.setColor(255, 255, 255, 255)
    for x = 1, blob.health do
        lg.draw(heartImg, (x - 1) * 70, 0, 0, 2, 2)
    end

    if worldState == "loss" or worldState == "win" then
        if worldState == "win" then
            local p = (lc.getTime() - stateChangeTime) / 1.7
            p = btk.inQuintic(p) * 255
            lg.setColor(255, 255, 255, p)
            lg.rectangle("fill", 0, 0, lg.getWidth(), lg.getHeight())
        end

        local message
        local color
        if worldState == "loss" then
            message = "You died..."
            color = 255
        else
            message = "You won!"
            color = 0
        end
        local width = bigFont:getWidth(message)
        
        lg.setColor(color, color, color, (lc.getTime() - stateChangeTime) / 3 * 255)

        lg.setFont(bigFont)
        lg.print(message, lg.getWidth() / 2 - width / 2, 100)

        width = bigFont:getWidth(replayButtonText)
        lg.print(replayButtonText, lg.getWidth() / 2 - width / 2, replayButtonY)
    end
end
